@@ -13,9 +13,8 @@ use num_traits::{AsPrimitive, PrimInt, Unsigned};
 
 use super::{
     dijkstra::dijkstra,
-    lex::Lexeme,
     parser::{AStackType, ParseRepair, Parser, Recoverer},
-    Span,
+    Lexeme, Span,
 };
 
 const PARSE_AT_LEAST: usize = 3; // N in Corchuelo et al.
@@ -100,16 +99,29 @@ impl<StorageT: PrimInt + Unsigned> PartialEq for PathFNode<StorageT> {
 
 impl<StorageT: PrimInt + Unsigned> Eq for PathFNode<StorageT> {}
 
-struct CPCTPlus<'a, 'b: 'a, 'input: 'b, StorageT: 'static + Eq + Hash, ActionT: 'a> {
-    parser: &'a Parser<'a, 'b, 'input, StorageT, ActionT>,
+struct CPCTPlus<
+    'a,
+    'b: 'a,
+    'input: 'b,
+    LexemeT: Lexeme<StorageT>,
+    StorageT: 'static + Eq + Hash,
+    ActionT: 'a,
+    ParamT: Copy,
+> {
+    parser: &'a Parser<'a, 'b, 'input, LexemeT, StorageT, ActionT, ParamT>,
 }
 
-pub(super) fn recoverer<'a, StorageT: 'static + Debug + Hash + PrimInt + Unsigned, ActionT: 'a>(
-    parser: &'a Parser<StorageT, ActionT>,
-) -> Box<dyn Recoverer<StorageT, ActionT> + 'a>
+pub(super) fn recoverer<
+    'a,
+    LexemeT: Lexeme<StorageT>,
+    StorageT: 'static + Debug + Hash + PrimInt + Unsigned,
+    ActionT: 'a,
+    ParamT: Copy,
+>(
+    parser: &'a Parser<LexemeT, StorageT, ActionT, ParamT>,
+) -> Box<dyn Recoverer<LexemeT, StorageT, ActionT, ParamT> + 'a>
 where
     usize: AsPrimitive<StorageT>,
-    u32: AsPrimitive<StorageT>,
 {
     Box::new(CPCTPlus { parser })
 }
@@ -118,22 +130,24 @@ impl<
         'a,
         'b: 'a,
         'input: 'b,
+        LexemeT: Lexeme<StorageT>,
         StorageT: 'static + Debug + Hash + PrimInt + Unsigned,
         ActionT: 'a,
-    > Recoverer<StorageT, ActionT> for CPCTPlus<'a, 'b, 'input, StorageT, ActionT>
+        ParamT: Copy,
+    > Recoverer<LexemeT, StorageT, ActionT, ParamT>
+    for CPCTPlus<'a, 'b, 'input, LexemeT, StorageT, ActionT, ParamT>
 where
     usize: AsPrimitive<StorageT>,
-    u32: AsPrimitive<StorageT>,
 {
     fn recover(
         &self,
         finish_by: Instant,
-        parser: &Parser<StorageT, ActionT>,
+        parser: &Parser<LexemeT, StorageT, ActionT, ParamT>,
         in_laidx: usize,
-        mut in_pstack: &mut Vec<StIdx>,
-        mut astack: &mut Vec<AStackType<ActionT, StorageT>>,
-        mut spans: &mut Vec<Span>,
-    ) -> (usize, Vec<Vec<ParseRepair<StorageT>>>) {
+        in_pstack: &mut Vec<StIdx>,
+        astack: &mut Vec<AStackType<LexemeT, ActionT>>,
+        spans: &mut Vec<Span>,
+    ) -> (usize, Vec<Vec<ParseRepair<LexemeT, StorageT>>>) {
         // This function implements a minor variant of the algorithm from "Repairing syntax errors
         // in LR parsers" by Rafael Corchuelo, Jose A. Perez, Antonio Ruiz, and Miguel Toro.
         //
@@ -237,9 +251,9 @@ where
         let laidx = apply_repairs(
             parser,
             in_laidx,
-            &mut in_pstack,
-            &mut Some(&mut astack),
-            &mut Some(&mut spans),
+            in_pstack,
+            &mut Some(astack),
+            &mut Some(spans),
             &rnk_rprs[0],
         );
 
@@ -251,12 +265,13 @@ impl<
         'a,
         'b: 'a,
         'input: 'b,
+        LexemeT: Lexeme<StorageT>,
         StorageT: 'static + Debug + Hash + PrimInt + Unsigned,
         ActionT: 'a,
-    > CPCTPlus<'a, 'b, 'input, StorageT, ActionT>
+        ParamT: Copy,
+    > CPCTPlus<'a, 'b, 'input, LexemeT, StorageT, ActionT, ParamT>
 where
     usize: AsPrimitive<StorageT>,
-    u32: AsPrimitive<StorageT>,
 {
     fn insert(&self, n: &PathFNode<StorageT>, nbrs: &mut Vec<(u16, PathFNode<StorageT>)>) {
         let laidx = n.laidx;
@@ -266,10 +281,10 @@ where
             }
 
             let next_lexeme = self.parser.next_lexeme(n.laidx);
-            let new_lexeme = Lexeme::new(
+            let new_lexeme = Lexeme::new_faulty(
                 StorageT::from(u32::from(tidx)).unwrap(),
                 next_lexeme.span().start(),
-                None,
+                0,
             );
             let (new_laidx, n_pstack) = self.parser.lr_cactus(
                 Some(new_lexeme),
@@ -360,7 +375,7 @@ where
         &self,
         in_laidx: usize,
         cnds: Vec<PathFNode<StorageT>>,
-    ) -> Vec<Vec<Vec<ParseRepair<StorageT>>>> {
+    ) -> Vec<Vec<Vec<ParseRepair<LexemeT, StorageT>>>> {
         fn traverse<StorageT: PrimInt>(
             rm: &Cactus<RepairMerge<StorageT>>,
         ) -> Vec<Vec<Repair<StorageT>>> {
@@ -414,7 +429,7 @@ where
         &self,
         mut laidx: usize,
         from: &[Repair<StorageT>],
-    ) -> Vec<ParseRepair<StorageT>> {
+    ) -> Vec<ParseRepair<LexemeT, StorageT>> {
         from.iter()
             .map(|y| match *y {
                 Repair::InsertTerm(token_idx) => ParseRepair::Insert(token_idx),
@@ -435,42 +450,39 @@ where
 
 /// Apply the `repairs` to `pstack` starting at position `laidx`: return the resulting parse
 /// distance and a new pstack.
-fn apply_repairs<'a, StorageT: 'static + Debug + Hash + PrimInt + Unsigned, ActionT: 'a>(
-    parser: &Parser<StorageT, ActionT>,
+fn apply_repairs<
+    'a,
+    LexemeT: Lexeme<StorageT>,
+    StorageT: 'static + Debug + Hash + PrimInt + Unsigned,
+    ActionT: 'a,
+    ParamT: Copy,
+>(
+    parser: &Parser<LexemeT, StorageT, ActionT, ParamT>,
     mut laidx: usize,
-    mut pstack: &mut Vec<StIdx>,
-    mut astack: &mut Option<&mut Vec<AStackType<ActionT, StorageT>>>,
-    mut spans: &mut Option<&mut Vec<Span>>,
-    repairs: &[ParseRepair<StorageT>],
+    pstack: &mut Vec<StIdx>,
+    astack: &mut Option<&mut Vec<AStackType<LexemeT, ActionT>>>,
+    spans: &mut Option<&mut Vec<Span>>,
+    repairs: &[ParseRepair<LexemeT, StorageT>],
 ) -> usize
 where
     usize: AsPrimitive<StorageT>,
-    u32: AsPrimitive<StorageT>,
 {
     for r in repairs.iter() {
         match *r {
             ParseRepair::Insert(tidx) => {
                 let next_lexeme = parser.next_lexeme(laidx);
-                let new_lexeme = Lexeme::new(
+                let new_lexeme = Lexeme::new_faulty(
                     StorageT::from(u32::from(tidx)).unwrap(),
                     next_lexeme.span().start(),
-                    None,
-                );
-                parser.lr_upto(
-                    Some(new_lexeme),
-                    laidx,
-                    laidx + 1,
-                    &mut pstack,
-                    &mut astack,
-                    &mut spans,
+                    0,
                 );
+                parser.lr_upto(Some(new_lexeme), laidx, laidx + 1, pstack, astack, spans);
             }
             ParseRepair::Delete(_) => {
                 laidx += 1;
             }
             ParseRepair::Shift(_) => {
-                laidx =
-                    parser.lr_upto(None, laidx, laidx + 1, &mut pstack, &mut astack, &mut spans);
+                laidx = parser.lr_upto(None, laidx, laidx + 1, pstack, astack, spans);
             }
         }
     }
@@ -478,9 +490,14 @@ where
 }
 
 /// Simplifies repair sequences, removes duplicates, and sorts them into order.
-fn simplify_repairs<StorageT: 'static + Hash + PrimInt + Unsigned, ActionT>(
-    parser: &Parser<StorageT, ActionT>,
-    all_rprs: &mut Vec<Vec<ParseRepair<StorageT>>>,
+fn simplify_repairs<
+    LexemeT: Lexeme<StorageT>,
+    StorageT: 'static + Hash + PrimInt + Unsigned,
+    ActionT,
+    ParamT: Copy,
+>(
+    parser: &Parser<LexemeT, StorageT, ActionT, ParamT>,
+    all_rprs: &mut Vec<Vec<ParseRepair<LexemeT, StorageT>>>,
 ) where
     usize: AsPrimitive<StorageT>,
 {
@@ -498,13 +515,13 @@ fn simplify_repairs<StorageT: 'static + Hash + PrimInt + Unsigned, ActionT>(
     // Use a HashSet as a quick way of deduplicating repair sequences: occasionally we can end up
     // with hundreds of thousands (!), and we don't have a sensible ordering on ParseRepair to make
     // it plausible to do a sort and dedup.
-    let mut hs: HashSet<Vec<ParseRepair<StorageT>>> = all_rprs.drain(..).collect();
+    let mut hs: HashSet<Vec<ParseRepair<LexemeT, StorageT>>> = all_rprs.drain(..).collect();
     all_rprs.extend(hs.drain());
 
     // Sort repair sequences:
     //   1) by whether they contain Inserts that are %insert_avoid
     //   2) by the number of repairs they contain
-    let contains_avoid_insert = |rprs: &Vec<ParseRepair<StorageT>>| -> bool {
+    let contains_avoid_insert = |rprs: &Vec<ParseRepair<LexemeT, StorageT>>| -> bool {
         for r in rprs.iter() {
             if let ParseRepair::Insert(tidx) = r {
                 if parser.grm.avoid_insert(*tidx) {
@@ -532,16 +549,21 @@ fn simplify_repairs<StorageT: 'static + Hash + PrimInt + Unsigned, ActionT>(
 /// `ParseRepair`s allow the same distance of parsing, then the `ParseRepair` which requires
 /// repairs over the shortest distance is preferred. Amongst `ParseRepair`s of the same rank, the
 /// ordering is non-deterministic.
-fn rank_cnds<'a, StorageT: 'static + Debug + Hash + PrimInt + Unsigned, ActionT: 'a>(
-    parser: &Parser<StorageT, ActionT>,
+fn rank_cnds<
+    'a,
+    LexemeT: Lexeme<StorageT>,
+    StorageT: 'static + Debug + Hash + PrimInt + Unsigned,
+    ActionT: 'a,
+    ParamT: Copy,
+>(
+    parser: &Parser<LexemeT, StorageT, ActionT, ParamT>,
     finish_by: Instant,
     in_laidx: usize,
     in_pstack: &[StIdx],
-    in_cnds: Vec<Vec<Vec<ParseRepair<StorageT>>>>,
-) -> Vec<Vec<ParseRepair<StorageT>>>
+    in_cnds: Vec<Vec<Vec<ParseRepair<LexemeT, StorageT>>>>,
+) -> Vec<Vec<ParseRepair<LexemeT, StorageT>>>
 where
     usize: AsPrimitive<StorageT>,
-    u32: AsPrimitive<StorageT>,
 {
     let mut cnds = Vec::new();
     let mut furthest = 0;
@@ -604,13 +626,13 @@ mod test {
     use num_traits::{AsPrimitive, PrimInt, ToPrimitive, Unsigned};
 
     use crate::{
-        lex::Lexeme,
-        parser::{test::do_parse, LexParseError, ParseRepair, RecoveryKind},
+        parser::{test::do_parse, ParseRepair, RecoveryKind},
+        LexParseError, Lexeme,
     };
 
-    fn pp_repairs<StorageT: 'static + Hash + PrimInt + Unsigned>(
+    fn pp_repairs<LexemeT: Lexeme<StorageT>, StorageT: 'static + Hash + PrimInt + Unsigned>(
         grm: &YaccGrammar<StorageT>,
-        repairs: &[ParseRepair<StorageT>],
+        repairs: &[ParseRepair<LexemeT, StorageT>],
     ) -> String
     where
         usize: AsPrimitive<StorageT>,
@@ -628,9 +650,12 @@ mod test {
         out.join(", ")
     }
 
-    fn check_all_repairs<StorageT: 'static + Debug + Hash + PrimInt + Unsigned>(
+    fn check_all_repairs<
+        LexemeT: Lexeme<StorageT>,
+        StorageT: 'static + Debug + Hash + PrimInt + Unsigned,
+    >(
         grm: &YaccGrammar<StorageT>,
-        err: &LexParseError<StorageT>,
+        err: &LexParseError<LexemeT, StorageT>,
         expected: &[&str],
     ) where
         usize: AsPrimitive<StorageT>,
@@ -718,7 +743,7 @@ E : 'N'
         let err_tok_id = u32::from(grm.token_idx("N").unwrap()).to_u16().unwrap();
         match &errs[0] {
             LexParseError::ParseError(e) => {
-                assert_eq!(e.lexeme(), &Lexeme::new(err_tok_id, 2, Some(1)))
+                assert_eq!(e.lexeme(), &Lexeme::new(err_tok_id, 2, 1))
             }
             _ => unreachable!(),
         }
